/*
 * This file is part of ascad.
 *
 * Ascad is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ascad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ascad; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
/* Copyright (C) 2005 Andrei Krger */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

#ifdef HAVE_LIBJPEG
#include <jpeglib.h>
#endif
#ifdef HAVE_LIBTIFF
#include <tiffio.h>
#endif
#ifdef HAVE_LIBPNG
#include <png.h>
#endif
#ifdef HAVE_LIBGIF
#include <gif_lib.h>
#endif

#include <setjmp.h>
#include "ascad.h"

#define return_err(e) {errstr=e; return -1;}

#ifdef HAVE_LIBJPEG
struct ascad_jpeg_error_mgr {
  struct jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
};

void jpeg_error_exit(j_common_ptr cinfo)
{
    struct ascad_jpeg_error_mgr * err = (struct ascad_jpeg_error_mgr *)cinfo->err;
    (*cinfo->err->output_message) (cinfo);
    longjmp(err->setjmp_buffer, 1);
}

int _LoadJPEG(FILE *fp, image_data_t *img)
{
  struct jpeg_decompress_struct cinfo;
  struct ascad_jpeg_error_mgr jerr;
  double dots_per_pixel;
  long   row_size;
  int    i;

  cinfo.err = jpeg_std_error(&jerr.pub);
  if (setjmp(jerr.setjmp_buffer))
  {
    jpeg_destroy_decompress(&cinfo);
    return_err("libjpeg error");
  }

  jpeg_create_decompress(&cinfo);

  jpeg_stdio_src(&cinfo, fp);
  jpeg_read_header(&cinfo, TRUE);

  cinfo.out_color_space = JCS_RGB;
  cinfo.do_fancy_upsampling = FALSE;
  cinfo.do_block_smoothing = FALSE;

  jpeg_start_decompress(&cinfo);

  assert(cinfo.output_components == 1);

  img->cols = cinfo.output_width;
  img->rows = cinfo.output_height;

  /* 1 = dots/inch, 2 = dots/cm, 0 = unknown */
  dots_per_pixel = (cinfo.density_unit == 2) ? 28.34 : 72.0;
  img->res_x = (double)cinfo.X_density / dots_per_pixel;
  img->res_y = (double)cinfo.Y_density / dots_per_pixel;

  row_size = img->cols * 3 * sizeof(byte);

  img->row_table = xmalloc(img->rows * sizeof(byte*));
  for(i=0; i<img->rows; i++)
    img->row_table[i] = xmalloc(row_size);

  while (cinfo.output_scanline < cinfo.output_height) 
    jpeg_read_scanlines(&cinfo, &img->row_table[cinfo.output_scanline], cinfo.output_height-cinfo.output_scanline);

  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  return 0;
}
#endif /* HAVE_LIBJPEG */

#ifdef HAVE_LIBPNG
int _LoadPNG(FILE *fp, image_data_t *img)
{
  png_struct *png;
  png_info *png_info;
  int color_type, bit_depth;
  long row_size;
  int i;

  if ((png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL)
    xerror("png_create_read_struct");

  if ((png_info = png_create_info_struct(png)) == NULL) 
    xerror("png_create_info_struct");

  if (setjmp(png_jmpbuf(png))) {
    png_destroy_read_struct(&png, &png_info, NULL);
    return_err("pnglib failed");
  }

  png_init_io(png, fp);

  png_read_info(png, png_info);

  img->cols = png_get_image_width(png, png_info);
  img->rows = png_get_image_height(png, png_info);
  img->res_x = (double)png_get_x_pixels_per_meter(png, png_info) / 2834;
  img->res_y = (double)png_get_y_pixels_per_meter(png, png_info) / 2834;

  color_type = png_get_color_type(png, png_info);
  bit_depth = png_get_bit_depth(png, png_info);

  png_set_strip_16(png);
  png_set_gray_to_rgb(png);
  png_set_expand(png);
  png_set_strip_alpha(png);
  png_set_packing(png);

  row_size = img->cols * 3 * sizeof(byte);

  img->row_table = xmalloc(img->rows * sizeof(byte*));
  for(i=0; i<img->rows; i++)
    img->row_table[i] = xmalloc(row_size);

  png_read_image(png, img->row_table);

  png_destroy_read_struct(&png, &png_info, NULL);

  return 0;

}
#endif /* HAVE_LIBPNG */

#ifdef HAVE_LIBTIFF

#define LINE_SIZE 1024

void ascad_TIFFErrorHandler(const char *module, const char *fmt, va_list ap) {
  errstr = xmalloc(LINE_SIZE);
  vsnprintf(errstr, LINE_SIZE, fmt, ap);
}

int _LoadTIFF(FILE *fp, image_data_t *img)
{
  TIFF               *tif;
  int                x, y, yr;
  dword              *data, *dptr, row_size;
  word               resol;
  float              xres, yres;

  TIFFSetErrorHandler(ascad_TIFFErrorHandler);
  TIFFSetWarningHandler(ascad_TIFFErrorHandler);

  fflush(fp);

  if ((tif = TIFFFdOpen(fileno(fp), "tiff", "r")) == NULL)
    return_err("tifflib failed");

  if (!TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &img->cols) ||
      !TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &img->rows))
    return_err("no height/width defined");

  if (TIFFGetField(tif, TIFFTAG_RESOLUTIONUNIT, &resol) &&
      TIFFGetField(tif, TIFFTAG_XRESOLUTION, &xres) &&
      TIFFGetField(tif, TIFFTAG_YRESOLUTION, &yres))
  {
    if (resol == RESUNIT_CENTIMETER)
      img->res_x = xres / 28.34, img->res_y = yres / 28.34;
    else if (resol == RESUNIT_INCH)
      img->res_x = xres / 72.0, img->res_y = yres / 72.0;
    else
      img->res_x = xres, img->res_y = yres;
  }
  else
    img->res_x = img->res_y = 1.0;

  row_size = img->cols * 3 * sizeof(byte);

  img->row_table = xmalloc(img->rows * sizeof(byte*));
  for(y=0; y<img->rows; y++)
    img->row_table[y] = xmalloc(row_size);

  if ((data = _TIFFmalloc(img->cols * img->rows * sizeof(dword))) == NULL)
    xerror("libtiff: out of memory");

  if (TIFFReadRGBAImage(tif, img->cols, img->rows, data, 0) == 0) {
    _TIFFfree(data);
    TIFFClose(tif);
    return_err("libtiff failed");
  }

  for (y = 0, yr=img->rows-1; y < img->rows; y++, yr--)
  {
    dptr = data + (yr * img->cols);
    for (x = 0; x < img->cols; x++)
    {
      img->row_table[y][x*3]   = TIFFGetR(dptr[x]);
      img->row_table[y][x*3+1] = TIFFGetB(dptr[x]);
      img->row_table[y][x*3+2] = TIFFGetG(dptr[x]);
    }
  }

  _TIFFfree(data);
  TIFFClose(tif);

  return 0;
}

#endif /* HAVE_LIBTIFF */

#ifdef HAVE_LIBGIF

#define HANDLE_ERROR {PrintGifError(); errstr = "libgif error"; return -1; }

int _LoadGIF(FILE *fp, image_data_t *img)
{
  GifFileType        *gif;
  GifRecordType       rec;
  ColorMapObject     *cmap;
  int csize;
  int                 i, j, done=0;
  long  row_size;
  int                 intoffset[] = {0, 4, 2, 1}, intjump[] = {8, 8, 4, 2};

  fflush(fp); 

  if ((gif = DGifOpenFileHandle(fileno(fp))) == NULL)
    HANDLE_ERROR;

  do
  {
    if (DGifGetRecordType(gif, &rec) == GIF_ERROR)
      HANDLE_ERROR;
    if ((rec == IMAGE_DESC_RECORD_TYPE) && (!done))
    {
      if (DGifGetImageDesc(gif) == GIF_ERROR)
        HANDLE_ERROR;

      img->cols = gif->Image.Width;
      img->rows = gif->Image.Height;
      img->res_x = img->res_y = 1.0;

      row_size = img->cols * 3 * sizeof(byte);
      img->row_table = xmalloc(img->rows * sizeof(byte*));

      for (i = 0; i < img->rows; i++)
        img->row_table[i] = xmalloc(row_size);

      if (gif->Image.Interlace)
      {
        for (j = 0; j < 4; j++)
          for (i = intoffset[j]; i < img->rows; i += intjump[j])
            DGifGetLine(gif, img->row_table[i], img->cols);
      }
      else
      {
        for (i = 0; i < img->rows; i++)
          DGifGetLine(gif, img->row_table[i], img->cols);
      }
      done = 1;
    }
  } while (rec != TERMINATE_RECORD_TYPE);

  cmap = (gif->Image.ColorMap ? gif->Image.ColorMap : gif->SColorMap);
  csize = cmap->ColorCount;

  for (i = 0; i < img->rows; i++)
    for (j = img->cols-1; j >= 0; j--)
    {
      byte id = img->row_table[i][j];
      img->row_table[i][j*3]   = cmap->Colors[id].Red;
      img->row_table[i][j*3+1] = cmap->Colors[id].Green;
      img->row_table[i][j*3+2] = cmap->Colors[id].Blue;
    }

  DGifCloseFile(gif);

  return 0;
}

#endif /* HAVE_LIBGIF */


#define BI_RGB       0
#define BI_RLE8      1
#define BI_RLE4      2
#define BI_BITFIELDS 3
#define LE_32(x) (*(dword*)(x))
#define LE_16(x) (*(word*)(x))
#define LE_8(x) (*(byte*)(x))

int _LoadBMP(FILE *fp, image_data_t *img)
{
  int    x, y, i;
  byte   buf[54];
  long   row_size;
  word   bpp;
  dword  compression, colors, offset;
  struct _cmap { byte r, g, b;} *cmap;

  /* reading the bmp header */

  if (fread(buf, 1, 54, fp) != 54)
    return_err("Error read BMP header");

  if (buf[0] != 'B' || buf[1] != 'M' || LE_32(buf+14) != 40)
    return_err("Unsupported BMP format");

  offset = LE_32(buf+10);
  bpp = LE_16(buf+28);
  compression = LE_32(buf+30);
  colors = LE_32(buf+46);
  img->cols = LE_32(buf+18);
  img->rows = LE_32(buf+22);
  img->res_x = (double)LE_32(buf+38) / 2834;
  img->res_y = (double)LE_32(buf+42) / 2834;

  if (bpp != 1 &&
      bpp != 4 &&
      bpp != 8 &&
      bpp != 16 &&
      bpp != 24 &&
      bpp != 32)
    return_err("bpp not supported");

  if (compression != BI_RGB &&
      compression != BI_RLE4 &&
      compression != BI_RLE8 /* &&
      compression != BI_BITFIELDS */)
    return_err("compression not supported");

  if (/*(compression == BI_RLE4 && bpp != 4) || */
      (compression == BI_RLE8 && bpp != 8) ||
      (compression == BI_BITFIELDS && (bpp != 16 && bpp != 32)))
    return_err("compression/bpp error");

  row_size = img->cols * 3 * sizeof(byte);

  img->row_table = xmalloc(img->rows * sizeof(byte*));
  for(i=0; i < img->rows; i++)
    img->row_table[i] = xmalloc(row_size);

  /* Reading the palette, if it exists.  */

  if (bpp < 16)
  {
    int n = (colors ? colors : 1 << bpp);
    cmap = (struct _cmap *)xmalloc(sizeof(struct _cmap) * n);
    for (i = 0; i < n; i++)
    {
      byte buf[4];
      fread(buf, 1, 4, fp);
      cmap[i].b = buf[0];
      cmap[i].g = buf[1];
      cmap[i].r = buf[2];
    }
  }
  else
    cmap = NULL;

  /* Reading the image data */

  if (fseek(fp, offset, SEEK_SET))
    return -1;

  if (bpp == 1)
  {
    for (y = img->rows-1; y >= 0; y--)
    {
      byte b=0, mask=0;
      int index;

      for (x = 0; x < img->cols; x++)
      {
        if (mask == 0)
          mask = 0x80, b = getc(fp);
        else 
          mask >>= 1;
        index = ((b & mask) ? 1 : 0);
        img->row_table[y][x*3]   = cmap[index].r;
        img->row_table[y][x*3+1] = cmap[index].g;
        img->row_table[y][x*3+2] = cmap[index].b;
      }
      for(x=(x+7)/8; x%4; x++) /* padding */
        getc(fp);
    }
  }
  else if (bpp == 4)
  {
    for (y = img->rows-1; y >= 0; y--)
    {
      byte b=0;
      int index, shift=0;

      for (x = 0; x < img->cols; x++)
      {
        if (shift == 0)
          b = getc(fp);
        shift = shift ? 0 : 4;
        index = (b >> shift) & 0xF;
        img->row_table[y][x*3]   = cmap[index].r;
        img->row_table[y][x*3+1] = cmap[index].g;
        img->row_table[y][x*3+2] = cmap[index].b;
      }
      for(x=(x+3)/4; x%4; x++) /* padding */
        getc(fp);
    }
  }
  else if (bpp == 8 && compression == BI_RLE8)
  {
    byte b, c;

    /* initialize background color */
    for (y = 0; y < img->rows; y++) 
      for (x = 0; x < img->cols; x++)
      {
        img->row_table[y][x] = cmap[0].r;
        img->row_table[y][x+1] = cmap[0].g;
        img->row_table[y][x+2] = cmap[0].b;
      }

    for (y = img->rows-1; y >= 0; y--)
    {
      for (x = 0; x < img->cols;)
      {

        if ((c = getc(fp)) == 0)
        {
          if ((c = getc(fp)) == 0) /* eoln */
          {
            x = img->cols;
          }
          else if (c == 1) /* eof */
          {
            x = img->cols;
            y = -1;
          }
          else if (c == 2)
          {
            x += getc(fp);
            y -= getc(fp);
          }
          else /* c > 2 */
          {
            for (i=0; i<c; i++)
            {
              b = getc(fp);
              img->row_table[y][x*3]   = cmap[b].r;
              img->row_table[y][x*3+1] = cmap[b].g;
              img->row_table[y][x*3+2] = cmap[b].b;
              x++;
            }
            if (c & 0x01) /* extra null for odd countres */
              b = getc(fp);
          }
        }
        else /* c != 0 */
        {
          b = getc(fp);
          for(i=0; i<c; i++)
          {
            img->row_table[y][x*3]   = cmap[b].r;
            img->row_table[y][x*3+1] = cmap[b].g;
            img->row_table[y][x*3+2] = cmap[b].b;
            x++;
          }
        }
      }
    }
  }
  else if (bpp == 8) /* && compression != BI_RLE8 */
  {
    for (y = img->rows-1; y >= 0; y--)
    {
      for (x = 0; x < img->cols; x++)
      {
        byte index = getc(fp);
        img->row_table[y][x*3]   = cmap[index].r;
        img->row_table[y][x*3+1] = cmap[index].g;
        img->row_table[y][x*3+2] = cmap[index].b;
      }
      for(; x%4; x++) /* padding */
        getc(fp);
    }
  }
  else if (bpp == 16)
  {
    for (y = img->rows-1; y >= 0; y--)
    {
      for (x = 0; x < img->cols; x++)
      {
        word w;
        fread(&w, 2, 1, fp);
        img->row_table[y][x*3]   = (w >> 10) & 0x1F;
        img->row_table[y][x*3+1] = (w >> 5) & 0x1F;
        img->row_table[y][x*3+2] = (w     )& 0x1F;
      }
      for(x *= 2; x%4; x++) /* padding */
        getc(fp);
    }
  }
  else if (bpp == 24)
  {
    for (y = img->rows-1; y >= 0; y--)
    {
      for (x = 0; x < img->cols; x++)
      {
        img->row_table[y][x*3+2] = getc(fp);
        img->row_table[y][x*3+1] = getc(fp);
        img->row_table[y][x*3]   = getc(fp);
      }
      for(x *= 3; x%4; x++) /* padding */
        getc(fp);
    }
  }
  else /* bpp == 32 */
  {
    for (y = img->rows-1; y >= 0; y--)
    {
      for (x = 0; x < img->cols; x++)
      {
        dword d;
        fread(&d, 4, 1, fp);

        img->row_table[y][x*3]   = (d >> 16) & 0xFF;
        img->row_table[y][x*3+1] = (d >> 8) & 0xFF;
        img->row_table[y][x*3+2] = (d    ) & 0xFF;
      }
    }
  }

  if (cmap)
    free(cmap);

  return 0;
}

int isjpeg(FILE *f)
{
  unsigned char       buf[8];

  if (!f)
    return 0;
  fread(buf, 1, 2, f);
  rewind(f);
  if ((buf[0] == 0xff) && (buf[1] == 0xd8))
    return 1;
  return 0;
}

#ifdef HAVE_LIBPNG
int ispng(FILE *f)
{
  unsigned char       buf[8];

  if (!f)
    return 0;
  fread(buf, 1, 8, f);
  rewind(f);
  return (int)png_check_sig(buf, 8);
}
#endif

int istiff(FILE *f)
{
  char                buf[8];

  if (!f)
    return 0;
  fgets(buf, 5, f);
  rewind(f);
  if ((buf[0] == 'M') && (buf[1] == 'M') && (buf[2] == 0x00) && (buf[3] == 0x2a))
    return 1;
  if ((buf[0] == 'I') && (buf[1] == 'I') && (buf[2] == 0x2a) && (buf[3] == 0x00))
    return 1;
  return 0;
}

int isgif(FILE *f)
{
  char                buf[8];

  if (!f)
    return 0;
  fread(buf, 1, 4, f);
  rewind(f);
  buf[4] = 0;
  if (!strcmp("GIF8", buf))
    return 1;
  return 0;
}

int isbmp(FILE *f)
{
  char                buf[3];

  if (!f)
    return 0;
  fread(buf, 1, 2, f);
  rewind(f);
  buf[2] = 0;
  if (!strcmp("BM", buf))
    return 1;
  return 0;
}

int load_image(FILE *fp, image_data_t *img)
{

  if (isbmp(fp))
    return _LoadBMP(fp, img);
    
#ifdef HAVE_LIBJPEG
  else if (isjpeg(fp))
    return _LoadJPEG(fp, img);
#endif

#ifdef HAVE_LIBTIFF
  else if (istiff(fp))
    return _LoadTIFF(fp, img);
#endif

#ifdef HAVE_LIBPNG
  else if (ispng(fp))
    return _LoadPNG(fp, img);
#endif

#ifdef HAVE_LIBGIF
  else if (isgif(fp))
    return _LoadGIF(fp, img);
#endif

  return_err("Image format not supported");

}


